package com.verifai.verifai_sdk

import android.graphics.Bitmap
import com.verifai.core.*
import com.verifai.core.filters.VerifaiDocumentBlackListFilter
import com.verifai.core.filters.VerifaiDocumentFilter
import com.verifai.core.filters.VerifaiDocumentTypeWhiteListFilter
import com.verifai.core.filters.VerifaiDocumentWhiteListFilter
import com.verifai.core.flow.views.uiComponents.imageCropper.Polygon
import com.verifai.core.internal.DocumentType
import com.verifai.core.result.*
import com.verifai.core.validators.*
import com.verifai.liveness.LivenessCheckStatus
import com.verifai.liveness.checks.CloseEyes
import com.verifai.liveness.checks.FaceMatching
import com.verifai.liveness.checks.Speech
import com.verifai.liveness.checks.Tilt
import com.verifai.liveness.result.VerifaiFaceMatchingCheckResult
import com.verifai.liveness.result.VerifaiLivenessCheckResults
import com.verifai.nfc.*
import com.verifai.nfc.mrtd.ReadStatus
import com.verifai.nfc.mrtd.LdsFile
import com.verifai.nfc.result.VerifaiNfcResult
import java.io.ByteArrayOutputStream
import java.security.cert.X509Certificate

fun Messages.MrzDataResult.convertFromMessageData(): VerifaiMrzDataResult {
    val result = this

    return VerifaiMrzDataResult(
        mrzString = result.mrzString,
        format = result.format,
        documentType = result.documentType,
        documentSubType = result.documentSubType,
        countryCode = result.countryCode,
        lastName = result.lastName,
        firstName = result.firstName,
        documentNumber = result.documentNumber,
        documentNumberCheckDigit = result.documentNumberCheckDigit,
        nationality = result.nationality,
        dateOfBirthRaw = result.dateOfBirthRaw,
        dateOfBirthCheckDigit = result.dateOfBirthCheckDigit,
        sex = result.sex,
        dateOfExpiryRaw = result.dateOfExpiryRaw,
        dateOfExpiryCheckDigit = result.dateOfExpiryCheckDigit,
        optionalData = result.optionalData,
        optionalDataCheckDigit = result.optionalDataCheckDigit,
        optionalDataTwo = result.optionalDataTwo,
        compositeCheckDigit = result.compositeCheckDigit,
        documentSpecificData = null,
        nfcKey = result.nfcKey,
        validityScore = result.validityScore?.toFloat(),
        isNfcKeyValid = result.isNfcKeyValid,
        isDocumentCodeValid = result.isDocumentCodeValid,
        isDocumentNumberValid = result.isDocumentNumberValid,
        isDateOfBirthValid = result.isDateOfBirthValid,
        isDateOfExpiryValid = result.isDateOfExpiryValid,
        isCompositeValid = result.isCompositeValid,
        isOptionalDataValid = result.isOptionalDataValid,
        isLine1LengthValid = result.isLine1LengthValid,
        isLine2LengthValid = result.isLine2LengthValid,
        isLine3LengthValid = result.isLine3LengthValid,
    )
}

fun VerifaiMrzDataResult.convertToMessageData(): Messages.MrzDataResult {
    val result = this

    val specData = result.documentSpecificData
    val docSpecificData = mutableMapOf<String, String?>()
    when (specData) {
        is FrenchIdData -> {
            docSpecificData["departmentOfIssuance"] = specData.departmentOfIssuance
            docSpecificData["officeOfIssuance"] = specData.officeOfIssuance
            docSpecificData["dateOfIssue"] = specData.dateOfIssue
        }
        is EdlData -> {
            docSpecificData["driversLicenceVersion"] = specData.driversLicenceVersion
            docSpecificData["driversLicenceRandomItem"] = specData.driversLicenceRandomItem
            docSpecificData["driversLicenceChip"] = specData.driversLicenceChip
        }
    }

    return Messages.MrzDataResult().apply {
        mrzString = result.mrzString
        format = result.format
        documentType = result.documentType
        documentSubType = result.documentSubType
        countryCode = result.countryCode
        lastName = result.lastName
        firstName = result.firstName
        documentNumber = result.documentNumber
        documentNumberCheckDigit = result.documentNumberCheckDigit
        nationality = result.nationality
        dateOfBirthRaw = result.dateOfBirthRaw
        dateOfBirthCheckDigit = result.dateOfBirthCheckDigit
        sex = result.sex
        dateOfExpiryRaw = result.dateOfExpiryRaw
        dateOfExpiryCheckDigit = result.dateOfExpiryCheckDigit
        optionalData = result.optionalData
        optionalDataCheckDigit = result.optionalDataCheckDigit
        optionalDataTwo = result.optionalDataTwo
        compositeCheckDigit = result.compositeCheckDigit
        documentSpecificData = docSpecificData
        nfcKey = result.nfcKey
        validityScore = result.validityScore?.toDouble()
        isNfcKeyValid = result.isNfcKeyValid
        isDocumentCodeValid = result.isDocumentCodeValid
        isDocumentNumberValid = result.isDocumentNumberValid
        isDateOfBirthValid = result.isDateOfBirthValid
        isDateOfExpiryValid = result.isDateOfExpiryValid
        isCompositeValid = result.isCompositeValid
        isOptionalDataValid = result.isOptionalDataValid
        isLine1LengthValid = result.isLine1LengthValid
        isLine2LengthValid = result.isLine2LengthValid
        isLine3LengthValid = result.isLine3LengthValid
    }
}

fun Polygon.convertToMessageData(): Messages.Polygon? {
    val postList = map {
        val builder = Messages.CoordinatePoint.Builder()
        builder.setX(it.x.toLong())
        builder.setY(it.y.toLong())
        builder.build()
    }

    if (postList.size != 4)
        return null

    return Messages.Polygon().apply {
        topLeft = postList[0]
        topRight = postList[1]
        bottomRight = postList[2]
        bottomLeft = postList[3]
    }
}

fun Bitmap.convertToMessageData(): ByteArray {
    val stream = ByteArrayOutputStream()
    compress(Bitmap.CompressFormat.PNG, 100, stream)
    return stream.toByteArray()
}

fun Messages.DocumentResult.convertFromMessageData(): VerifaiDocumentResult {
    val result = this;
    return VerifaiDocumentResult(
        uuid = result.uuid,
        type = result.type,
        model = result.model,
        country = result.country,
        widthMm = result.widthMm?.toBigDecimal(),
        heightMm = result.heightMm?.toBigDecimal(),
        sampleFront = result.sampleFront,
        mrzType = result.mrzType,
        mrzStart = result.mrzStart,
        nfcType = result.nfcType,
        zones = result.zones!!.map {
            VerifaiDocumentZoneResult(
                title = it.title!!,
                side = it.side!!,
                x = it.x!!.toBigDecimal(),
                y = it.y!!.toBigDecimal(),
                width = it.width!!.toBigDecimal(),
                height = it.height!!.toBigDecimal(),
                block = it.block!!,
            )
        },
        numberOfSides = result.numberOfSides?.toInt(),
    )
}

fun VerifaiDocumentResult.convertToMessageData(): Messages.DocumentResult {
    val result = this;
    return Messages.DocumentResult().apply {
        uuid = result.uuid
        type = result.type
        model = result.model
        country = result.country
        widthMm = result.widthMm?.toDouble()
        heightMm = result.heightMm?.toDouble()
        sampleFront = result.sampleFront
        mrzType = result.mrzType
        mrzStart = result.mrzStart
        nfcType = result.nfcType
        zones = result.zones.map {
            Messages.DocumentZoneResult().apply {
                title = it.title
                side = it.side
                x = it.x.toDouble()
                y = it.y.toDouble()
                width = it.width.toDouble()
                height = it.height.toDouble()
                block = it.block
            }
        }
        numberOfSides = result.numberOfSides?.toLong()
        hasMrz = result.hasMrz
    }
}

fun VerifaiBarcodeSymbology.convertToMessageData(): Messages.BarcodeSymbology {
    return when (this) {
        VerifaiBarcodeSymbology.QR -> Messages.BarcodeSymbology.QR
        VerifaiBarcodeSymbology.CODE128 -> Messages.BarcodeSymbology.CODE128
        VerifaiBarcodeSymbology.CODE39 -> Messages.BarcodeSymbology.CODE39
        VerifaiBarcodeSymbology.PDF417 -> Messages.BarcodeSymbology.PDF417
    }
}

fun VerifaiDocumentSide.convertToMessageData(): Messages.DocumentSide {
    return when (this) {
        VerifaiDocumentSide.BACK -> Messages.DocumentSide.BACK
        VerifaiDocumentSide.FRONT -> Messages.DocumentSide.FRONT
    }
}

fun VerifaiBarcodeResult.convertToMessageData(): Messages.BarcodeResult {
    val result = this
    return Messages.BarcodeResult().apply {
        symbology = result.symbology.convertToMessageData()
        side = result.side.convertToMessageData()
        data = result.data
    }
}

fun VerifaiResult.convertToMessageData(): Messages.CoreResult {
    val result = this
    return Messages.CoreResult().apply {
        frontImage = result.frontImage?.convertToMessageData()
        backImage = result.backImage?.convertToMessageData()
        mrzData = result.mrzData?.convertToMessageData()
        frontDocumentPosition = result.frontDocumentPosition?.convertToMessageData()
        backDocumentPosition = result.backDocumentPosition?.convertToMessageData()
        document = result.document?.convertToMessageData()
        documentBarcodes = result.documentBarcodes?.map { it.convertToMessageData() }
        visualInspectionZoneResult = result.visualInspectionZoneResult
    }
}

fun Messages.Validator.convertFromMessageData(): VerifaiValidatorInterface {
    when (type) {
        Messages.ValidatorType.COUNTRY_ALLOWLIST -> {
            args?.let {
                if (it.contains("countryList")) {
                    @Suppress("UNCHECKED_CAST")
                    return VerifaiDocumentCountryWhiteListValidator(
                        null,
                        it["countryList"] as MutableList<String>
                    )
                } else {
                    throw Exception("Allowlist validator should have 'countryList' argument")
                }
            } ?: let {
                throw Exception("Allowlist validator should have 'args' argument")
            }
        }
        Messages.ValidatorType.COUNTRY_BLOCKLIST -> {
            args?.let {
                if (it.contains("countryList")) {
                    @Suppress("UNCHECKED_CAST")
                    return VerifaiDocumentCountryBlackListValidator(
                        null,
                        it["countryList"] as MutableList<String>
                    )
                } else {
                    throw Exception("Blocklist validator should have 'countryList' argument")
                }
            } ?: let {
                throw Exception("Blocklist validator should have 'args' argument")
            }
        }
        Messages.ValidatorType.HAS_MRZ -> {
            return VerifaiDocumentHasMrzValidator()
        }
        Messages.ValidatorType.DOCUMENT_TYPES -> {
            args?.let {
                if (it.contains("validDocumentTypes")) {
                    return VerifaiDocumentTypesValidator(null, null, createDocumentTypes(it))
                } else {
                    throw Exception("DocumentTypes validator should have 'validDocumentTypes' argument")
                }
            } ?: run {
                throw Exception("DocumentTypes  validator should have 'args' argument")
            }
        }
        Messages.ValidatorType.MRZ_AVAILABLE -> {
            return VerifaiMrzAvailableValidator(null)
        }
        Messages.ValidatorType.NFC_KEY_WHEN_AVAILABLE -> {
            return VerifaiNFCKeyWhenAvailableValidator()
        }
        else -> {
            throw Exception("Wrong argument type")
        }
    }
}

fun createDocumentTypes(docTypes: Map<String, Any>): ArrayList<DocumentType> {
    val documentTypes = arrayListOf<DocumentType>()
    @Suppress("UNCHECKED_CAST")
    for (doc in docTypes["validDocumentTypes"] as List<Int>) {
        when (doc) {
            Messages.DocumentType.ID_CARD.ordinal -> {
                documentTypes.add(DocumentType.IDENTITY_CARD)
            }
            Messages.DocumentType.DRIVERS_LICENCE.ordinal -> {
                documentTypes.add(DocumentType.DRIVING_LICENCE)
            }
            Messages.DocumentType.PASSPORT.ordinal -> {
                documentTypes.add(DocumentType.PASSPORT)
            }
            Messages.DocumentType.REFUGEE.ordinal -> {
                documentTypes.add(DocumentType.REFUGEE_TRAVEL_DOCUMENT)
            }
            Messages.DocumentType.EMERGENCY_PASSPORT.ordinal -> {
                documentTypes.add(DocumentType.EMERGENCY_PASSPORT)
            }
            Messages.DocumentType.RESIDENCE_PERMIT_TYPE_I.ordinal -> {
                documentTypes.add(DocumentType.RESIDENCE_PERMIT_I)
            }
            Messages.DocumentType.RESIDENCE_PERMIT_TYPE_II.ordinal -> {
                documentTypes.add(DocumentType.RESIDENCE_PERMIT_II)
            }
            Messages.DocumentType.VISA.ordinal -> {
                documentTypes.add(DocumentType.VISA)
            }
            Messages.DocumentType.UNKNOWN.ordinal -> {
                documentTypes.add(DocumentType.UNKNOWN)
            }
            else -> {
                throw Exception("Unknown document type")
            }
        }
    }
    return documentTypes
}

fun Messages.DocumentFilter.convertFromMessageData(): VerifaiDocumentFilter {
    when (type) {
        Messages.DocumentFilterType.DOCUMENT_ALLOW_LIST -> {
            args?.let {
                if (it.contains("countryList")) {
                    @Suppress("UNCHECKED_CAST")
                    return VerifaiDocumentWhiteListFilter(
                        it["countryList"] as ArrayList<String>
                    )
                } else {
                    throw Exception("Allowlist filter type should have 'countryList' argument")
                }
            } ?: run {
                throw Exception("Allowlist filter type should have 'args' argument")
            }
        }
        Messages.DocumentFilterType.DOCUMENT_BLOCK_LIST -> {
            args?.let {
                if (it.contains("countryList")) {
                    @Suppress("UNCHECKED_CAST")
                    return VerifaiDocumentBlackListFilter(
                        it["countryList"] as ArrayList<String>
                    )
                } else {
                    throw Exception("Allowlist filter type should have 'countryList' argument")
                }
            } ?: run {
                throw Exception("Allowlist filter type should have 'args' argument")
            }
        }
        Messages.DocumentFilterType.DOCUMENT_TYPE_ALLOW_LIST -> {
            args?.let {
                if (it.contains("validDocumentTypes")) {
                    return VerifaiDocumentTypeWhiteListFilter(createDocumentTypes(it))
                } else {
                    throw Exception("DocumentTypes filter should have 'validDocumentTypes' argument")
                }
            } ?: run {
                throw Exception("DocumentTypes  validator should have 'args' argument")
            }
        }
        else -> {
            throw Exception("Wrong argument type")
        }
    }
}

fun Messages.InstructionScreenConfiguration.convertFromMessageData(): VerifaiInstructionScreenConfiguration {
    val screenConfig = VerifaiInstructionScreenConfiguration()
    showInstructionScreens?.let { screenConfig.showInstructionScreens = it }
    instructionScreens?.let {
        val screens = mutableMapOf<VerifaiInstructionScreenId, VerifaiSingleInstructionScreen>()
        screenConfig.instructionScreens = screens
    }

    return screenConfig
}

fun Messages.Configuration.convertFromMessageData(): VerifaiConfiguration {
    val convertedValidators = validators?.map { it.convertFromMessageData() }
    val convertedFilters = documentFilters?.map { it.convertFromMessageData() }
    return VerifaiConfiguration(
        require_document_copy = requireDocumentCopy ?: true,
        enable_post_cropping = enablePostCropping ?: true,
        enable_manual = enableManual ?: true,
        require_mrz_contents = requireMrzContents ?: false,
        require_nfc_when_available = requireNfcWhenAvailable ?: false,
        read_mrz_contents = readMrzContents ?: true,
        extraValidators = convertedValidators ?: emptyList(),
        document_filters = convertedFilters ?: emptyList(),
        document_filters_auto_create_validators = documentFiltersAutoCreateValidators ?: true,
        is_scan_help_enabled = isScanHelpEnabled ?: true,
        require_cropped_image = requireCroppedImage ?: true,
        enableVisualInspection = enableVisualInspection ?: false,
        instructionScreenConfiguration = instructionScreenConfiguration?.convertFromMessageData()
            ?: VerifaiInstructionScreenConfiguration(),
    )
}

fun LivenessCheckStatus.convertToMessageData(): Messages.LivenessCheckStatus {
    return when (this) {
        LivenessCheckStatus.SUCCESS -> Messages.LivenessCheckStatus.SUCCESS
        LivenessCheckStatus.FAILED -> Messages.LivenessCheckStatus.FAILED
        LivenessCheckStatus.UNDETERMINED -> Messages.LivenessCheckStatus.UNDETERMINED
        LivenessCheckStatus.NOT_STARTED -> Messages.LivenessCheckStatus.NOT_STARTED
        LivenessCheckStatus.SKIPPED -> Messages.LivenessCheckStatus.SKIPPED
    }
}

fun VerifaiLivenessCheckResults.convertToMessageData(): Messages.LivenessResult {
    val result = this

    return Messages.LivenessResult().apply {
        automaticChecksPassed = result.automaticChecksPassed
        successRatio = result.successRatio.toDouble()
        resultList = result.resultList.map {
            val specificData = mutableMapOf<String, Any>()
            var checkName = ""
            when (it.check) {
                is CloseEyes -> {
                    checkName = "CloseEyes"
                    specificData["numberOfSeconds"] = (it.check as CloseEyes).numberOfSeconds
                }
                is Tilt -> {
                    checkName = "Tilt"
                    specificData["faceAngleRequirement"] = (it.check as Tilt).faceAngleRequirement
                }
                is Speech -> {
                    checkName = "Speech"
                    specificData["speechRequirement"] = (it.check as Speech).speechRequirement
                }
                is FaceMatching -> {
                    checkName = "faceMatching"
                    if (it is VerifaiFaceMatchingCheckResult) {
                        it.confidence?.let { c -> specificData["confidence"] = c }
                        it.match?.let { m -> specificData["match"] = m }
                    }
                }
            }
            Messages.LivenessCheckResult().apply {
                name = checkName
                instruction = it.check.instruction
                status = it.status.convertToMessageData()
                checkSpecificData = specificData
                recordingFileName = it.recording?.absolutePath
            }
        }
    }
}

fun BacStatus.convertToMessageData(): Messages.BacStatus {
    return when (this) {
        BacStatus.FAILED -> Messages.BacStatus.FAILED
        BacStatus.NOT_CHECKED -> Messages.BacStatus.NOT_CHECKED
        BacStatus.RANDOM_FAILED -> Messages.BacStatus.RANDOM_FAILED
        BacStatus.SUCCESS -> Messages.BacStatus.SUCCESS
    }
}

fun PaceStatus.convertToMessageData(): Messages.PaceStatus {
    return when (this) {
        PaceStatus.FAILED -> Messages.PaceStatus.FAILED
        PaceStatus.NOT_CHECKED -> Messages.PaceStatus.NOT_CHECKED
        PaceStatus.NOT_SUPPORTED -> Messages.PaceStatus.NOT_SUPPORTED
        PaceStatus.SUCCESS -> Messages.PaceStatus.SUCCESS
    }
}

fun AaStatus.convertToMessageData(): Messages.AaStatus {
    return when (this) {
        AaStatus.FAILED -> Messages.AaStatus.FAILED
        AaStatus.NOT_CHECKED -> Messages.AaStatus.NOT_CHECKED
        AaStatus.SIGNATURE_FAILED -> Messages.AaStatus.SIGNATURE_FAILED
        AaStatus.SUCCESS -> Messages.AaStatus.SUCCESS
        AaStatus.NOT_PRESENT -> Messages.AaStatus.NOT_PRESENT
    }
}

fun RootCertificateStatus.convertToMessageData(): Messages.RootCertificateStatus {
    return when (this) {
        RootCertificateStatus.INVALID -> Messages.RootCertificateStatus.INVALID
        RootCertificateStatus.NOT_CHECKED -> Messages.RootCertificateStatus.NOT_CHECKED
        RootCertificateStatus.NOT_FOUND -> Messages.RootCertificateStatus.NOT_FOUND
        RootCertificateStatus.REVOKED -> Messages.RootCertificateStatus.REVOKED
        RootCertificateStatus.VALID -> Messages.RootCertificateStatus.VALID
    }
}

fun CaStatus.convertToMessageData(): Messages.CaStatus {
    return when (this) {
        CaStatus.FAILED -> Messages.CaStatus.FAILED
        CaStatus.NOT_CHECKED -> Messages.CaStatus.NOT_CHECKED
        CaStatus.NOT_PRESENT -> Messages.CaStatus.NOT_PRESENT
        CaStatus.SUCCESS -> Messages.CaStatus.SUCCESS
    }
}

fun ReadStatus.convertToMessageData(): Messages.ReadStatus {
    return when (this) {
        ReadStatus.FAILED -> Messages.ReadStatus.FAILED
        ReadStatus.FILE_NOT_FOUND -> Messages.ReadStatus.FILE_NOT_FOUND
        ReadStatus.NO_ACCESS -> Messages.ReadStatus.NO_ACCESS
        ReadStatus.SUCCESS -> Messages.ReadStatus.SUCCESS
    }
}

fun LdsFile.convertToMessageData(): Messages.LdsFile {
    val result = this
    return Messages.LdsFile().apply {
        data = result.data
        dataHash = result.hash
        readStatus = result.readStatus.convertToMessageData()
    }
}

fun X509Certificate.convertToMessageData(): Messages.X509Certificate {
    val result = this
    return Messages.X509Certificate().apply {
        version = result.version.toLong()
        serialNumber = result.serialNumber.toLong()
        signatureAlgorithmOID = result.sigAlgOID
        signatureAlgorithmName = result.sigAlgName
        signatureAlgorithmParams = result.sigAlgParams
        issuerName = result.issuerX500Principal.name
        notBefore = result.notBefore.toString()
        notAfter = result.notAfter.toString()
        subjectName = result.subjectX500Principal.name
        tbsCertificate = result.tbsCertificate
        signature = result.signature
        issuerUniqueID = result.issuerUniqueID?.toList()
        subjectUniqueID = result.subjectUniqueID?.toList()
        keyUsage = result.keyUsage?.toList()
        extendedKeyUsage = result.extendedKeyUsage
    }
}

fun VerifaiNfcResult.convertToMessageData(): Messages.NfcResult {
    val result = this
    return Messages.NfcResult().apply {
        originality = result.originality()
        authenticity = result.authenticity()
        confidentiality = result.confidentiality()
        mrzMatch = result.mrzMatch
        comSodMatch = result.comSodMatch
        paceStatus = result.paceStatus.convertToMessageData()
        bacStatus = result.bacStatus.convertToMessageData()
        activeAuthenticationStatus = result.activeAuthenticationStatus.convertToMessageData()
        documentCertificateValid = result.documentCertificateValid
        signingCertificateMatchesWithParent =
            result.signingCertificateMatchesWithParent.convertToMessageData()
        scanCompleted = result.scanCompleted
        chipAuthenticationStatus = result.chipAuthenticationStatus.convertToMessageData()
        documentSignatureCorrect = result.documentSignatureCorrect
        photo = result.photo?.convertToMessageData()
        mrzData = result.mrzData?.convertToMessageData()
        dataGroups = result.dataGroups?.entries?.associate {
            it.key.toLong() to it.value.convertToMessageData()
        }
        documentCertificate = result.documentCertificate?.convertToMessageData()
        sodHashes = result.sodHashes?.entries?.associate {
            it.key.toLong() to it.value
        }
        sodData = result.sodData
        featurePoints = result.featurePoints?.toLong()
        mimeBytes = result.mimeBytes
        aaDigestAlgorithm = result.aaDigestAlgorithm
        chipAuthenticationOid = result.chipAuthenticationOid
        chipAuthenticationAgreementAlgorithm = result.chipAuthenticationAgreementAlgorithm
        chipAuthenticationPublicKeyAlgorithm = result.chipAuthenticationPublicKeyAlgorithm
        signingCertificate = result.signingCertificate?.convertToMessageData()
    }
}